package dev.peerat.framework;

import java.lang.reflect.Method;
import java.net.ServerSocket;
import java.net.Socket;
import java.security.MessageDigest;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.net.ssl.SSLServerSocket;
import javax.net.ssl.SSLServerSocketFactory;

import org.jose4j.jwa.AlgorithmConstraints.ConstraintType;
import org.jose4j.jwk.RsaJsonWebKey;
import org.jose4j.jwk.RsaJwkGenerator;
import org.jose4j.jws.AlgorithmIdentifiers;
import org.jose4j.jws.JsonWebSignature;
import org.jose4j.jwt.JwtClaims;
import org.jose4j.jwt.consumer.JwtConsumer;
import org.jose4j.jwt.consumer.JwtConsumerBuilder;
import org.jose4j.lang.JoseException;

public class Router<U extends User>{
	
	public static void main(String[] args) {
		
	}
	
	private Locker<Context> logger;
	private Map<Response, Route>[] responses;
	private Map<Response, Pattern> patterns;
	private Response noFileFound;
	private RsaJsonWebKey rsaJsonWebKey;
	private JwtConsumer jwtConsumer;
	private Consumer<JwtClaims> claims;
	private Function<JwtClaims, U> userCreator;
	private String[][] headers;
	private ServerSocket serverSocket;

	public Router() throws Exception{
		this.logger = new Locker<>();
		int types = RequestType.values().length;
		this.responses = new HashMap[types];
		for(RequestType type : RequestType.values()) this.responses[type.ordinal()] = new HashMap<>();
		this.patterns = new HashMap<>();
		this.rsaJsonWebKey = RsaJwkGenerator.generateJwk(2048);
		this.headers = new String[types][0];
	}
	
	public Router<U> configureJwt(Consumer<JwtConsumerBuilder> consumer, Consumer<JwtClaims> claims, Function<JwtClaims, U> userCreator){
		JwtConsumerBuilder builder = new JwtConsumerBuilder()
		.setRequireExpirationTime()
		.setAllowedClockSkewInSeconds(30)
		.setVerificationKey(rsaJsonWebKey.getKey())
		.setJwsAlgorithmConstraints(ConstraintType.PERMIT, AlgorithmIdentifiers.RSA_USING_SHA256);
		
		consumer.accept(builder);
		
		this.jwtConsumer = builder.build();
		this.claims = claims;
		this.userCreator = userCreator;
		return this;
	}
	
	public void addDefaultHeaders(RequestType type, String... headers){
		String[] origin = this.headers[type.ordinal()]; 
		String[] copy = new String[origin.length+headers.length];
		System.arraycopy(origin, 0, copy, 0, origin.length);
		System.arraycopy(headers, 0, copy, origin.length, headers.length);
		this.headers[type.ordinal()] = copy;
	}
	
	public void listen(int port, boolean ssl) throws Exception{
		if (ssl) { // Not needed with the use of a proxy
			try {
				SSLServerSocketFactory ssf = (SSLServerSocketFactory) SSLServerSocketFactory.getDefault();
				serverSocket = (SSLServerSocket) ssf.createServerSocket(port);

				while (!serverSocket.isClosed()) {
					Socket socket = serverSocket.accept();
					Client<U> client = new Client<>(socket, this);
					client.start();
				}
			} catch (Exception e) {
				e.printStackTrace();
			} finally {
				stop();
			}
		} else {
			try{
				serverSocket = new ServerSocket(port);
				while (!serverSocket.isClosed()) {
					Socket socket = serverSocket.accept();
					Client<U> client = new Client<>(socket, this);
					client.start();
				}
			} catch (Exception e) {
				e.printStackTrace();
			} finally {
				stop();
			}
		}
	}
	
	
	
	public void stop(){
		if(serverSocket == null) return;
		try {
			serverSocket.close();
			serverSocket = null;
		}catch(Exception e){
			e.printStackTrace();
		}
	}

	public Router<U> register(Response response){
		try{
			Method method = response.getClass().getDeclaredMethod("exec",
					Response.class.getDeclaredMethods()[0].getParameterTypes());
			Route route = method.getAnnotation(Route.class);

			this.responses[route.type().ordinal()].put(response, route);
			this.patterns.put(response, Pattern.compile(route.path()));
		}catch(Exception e){
			throw new IllegalArgumentException(e);
		}
		return this;
	}

	public Router<U> setDefault(Response response){
		this.noFileFound = response;
		return this;
	}

	Context exec(RequestType type, String path, User user, HttpReader reader, HttpWriter writer) throws Exception{
		if(type == null) return null;
		Context context = new Context(type, path, user, writer, this.headers[type.ordinal()]);
		for(Entry<Response, Route> routes : this.responses[type.ordinal()].entrySet()){
			Matcher matcher = this.patterns.get(routes.getKey()).matcher(path);
			if(matcher.matches()){
				if(user == null && routes.getValue().needLogin()){
					writer.response(401, "Access-Control-Allow-Origin: *");
					return context;
				}
				if(routes.getValue().websocket()){
					switchToWebSocket(reader, writer);
					reader = new WebSocketReader(reader);
					writer = new WebSocketWriter(writer);
				}
				routes.getKey().exec(matcher, context, reader, writer);
				return context;
			}
		}
		if(noFileFound != null) noFileFound.exec(null, context, reader, writer);
		return context;
	}

	public Router<U> configureSSL(String keyStore, String keyStorePassword){
		System.setProperty("javax.net.ssl.keyStore", keyStore);
		System.setProperty("javax.net.ssl.keyStorePassword", keyStorePassword);
		return this;
	}
	
	public U getUser(String token) throws Exception{
		return this.userCreator.apply(this.jwtConsumer.processToClaims(token));
	}

	public String createAuthUser(U user) throws JoseException{
		JwtClaims claims = new JwtClaims();
		claims.setGeneratedJwtId(); // a unique identifier for the token
		claims.setIssuedAtToNow(); // when the token was issued/created (now)
		claims.setNotBeforeMinutesInThePast(2); // time before which the token is not yet valid (2 minutes ago)
		this.claims.accept(claims);

		user.write(claims);

		JsonWebSignature jws = new JsonWebSignature();
		jws.setPayload(claims.toJson());
		jws.setKey(rsaJsonWebKey.getPrivateKey());
		jws.setKeyIdHeaderValue(rsaJsonWebKey.getKeyId());
		jws.setAlgorithmHeaderValue(AlgorithmIdentifiers.RSA_USING_SHA256);
		return jws.getCompactSerialization();
	}
	
	public Locker<Context> getLogger(){
		return this.logger;
	}
	
	private void switchToWebSocket(HttpReader reader, HttpWriter writer) throws Exception{
		String key = reader.getHeader("Sec-WebSocket-Key");
		if (key == null) throw new IllegalArgumentException();

		writer.write("HTTP/1.1 101 Switching Protocols\n");
		writer.write("Connection: Upgrade\n");
		writer.write("Upgrade: websocket\n");
		writer.write("Sec-WebSocket-Accept: " + printBase64Binary(MessageDigest.getInstance("SHA-1")
					.digest((key + "258EAFA5-E914-47DA-95CA-C5AB0DC85B11").getBytes("UTF-8"))) + "\n");
		writer.write("\n");
		writer.flush();
	}


	// From javax.xml.bind.DatatypeConverter
	private String printBase64Binary(byte[] array){
		char[] arrayOfChar = new char[(array.length + 2) / 3 * 4];
		int i = _printBase64Binary(array, 0, array.length, arrayOfChar, 0);
		assert i == arrayOfChar.length;
		return new String(arrayOfChar);
	}

	private int _printBase64Binary(byte[] paramArrayOfbyte, int paramInt1, int paramInt2,
			char[] paramArrayOfchar, int paramInt3){
		int i = paramInt2;
		int j;
		for (j = paramInt1; i >= 3; j += 3){
			paramArrayOfchar[paramInt3++] = encode(paramArrayOfbyte[j] >> 2);
			paramArrayOfchar[paramInt3++] = encode(
					(paramArrayOfbyte[j] & 0x3) << 4 | paramArrayOfbyte[j + 1] >> 4 & 0xF);
			paramArrayOfchar[paramInt3++] = encode(
					(paramArrayOfbyte[j + 1] & 0xF) << 2 | paramArrayOfbyte[j + 2] >> 6 & 0x3);
			paramArrayOfchar[paramInt3++] = encode(paramArrayOfbyte[j + 2] & 0x3F);
			i -= 3;
		}
		if (i == 1){
			paramArrayOfchar[paramInt3++] = encode(paramArrayOfbyte[j] >> 2);
			paramArrayOfchar[paramInt3++] = encode((paramArrayOfbyte[j] & 0x3) << 4);
			paramArrayOfchar[paramInt3++] = '=';
			paramArrayOfchar[paramInt3++] = '=';
		}
		if (i == 2){
			paramArrayOfchar[paramInt3++] = encode(paramArrayOfbyte[j] >> 2);
			paramArrayOfchar[paramInt3++] = encode(
					(paramArrayOfbyte[j] & 0x3) << 4 | paramArrayOfbyte[j + 1] >> 4 & 0xF);
			paramArrayOfchar[paramInt3++] = encode((paramArrayOfbyte[j + 1] & 0xF) << 2);
			paramArrayOfchar[paramInt3++] = '=';
		}
		return paramInt3;
	}

	private char encode(int paramInt){
		return encodeMap[paramInt & 0x3F];
	}

	private static final char[] encodeMap = initEncodeMap();

	private static char[] initEncodeMap(){
		char[] arrayOfChar = new char[64];
		byte b;
		for (b = 0; b < 26; b++)
			arrayOfChar[b] = (char) (65 + b);
		for (b = 26; b < 52; b++)
			arrayOfChar[b] = (char) (97 + b - 26);
		for (b = 52; b < 62; b++)
			arrayOfChar[b] = (char) (48 + b - 52);
		arrayOfChar[62] = '+';
		arrayOfChar[63] = '/';
		return arrayOfChar;
	}
}